#import "AudioFilePlayerPatch.h"

@implementation AudioFilePlayerPatch : QCPatch

+ (int)executionModeWithIdentifier:(id)fp8
{
	return 1;
}
+ (BOOL)allowsSubpatchesWithIdentifier:(id)fp8
{
	return NO;
}
+ (int)timeModeWithIdentifier:(id)fp8
{
	return 0;
}

- (id)attributes
{
	NSMutableDictionary *at=[[super attributes] mutableCopy];
	[at setObject:@"Kineme Audio File Player" forKey:@"name"];
	return at;
}


- (id)initWithIdentifier:(id)fp8
{
	self=[super initWithIdentifier:fp8];
	if(self)
	{
		[inputCurrentVolume setDoubleValue:1.0];
		[inputCurrentVolume setMinDoubleValue:0.0];
		[inputCurrentVolume setMaxDoubleValue:1.0];

		[inputCurrentPosition setMinDoubleValue:0.0];
	}
	return self;
}

- (BOOL)setup:(QCOpenGLContext *)context
{
	_allocatedSounds=[[NSMutableArray alloc] initWithCapacity:16];
	_executedSinceSetup=NO;

	return YES;
}
- (void)cleanup:(QCOpenGLContext *)context
{
	[_allocatedSounds release];
}

- (void)enable:(QCOpenGLContext *)context
{
	// if we've already executed, and the trigger was already on upon enable, launch playback (useful when switching between windowed-mode and fullscreen-mode)
	if( _executedSinceSetup && ![inputTrig wasUpdated] && [inputTrig booleanValue] )
		[self _startPlaying];
}

- (void)disable:(QCOpenGLContext *)context
{
	// stop any remaining sounds (since releasing the NSSound doesn't stop it)
	for(NSSound *s in _allocatedSounds)
	{
		// stop receiving events (we don't want to double-deallocate)
		[s setDelegate:nil];
		[s stop];
//		NSLog(@"AudioFilePlayerPatch: %08x: release  (disable)",s);
	}

	// this automatically releases all of the child NSSound objects
	[_allocatedSounds removeAllObjects];
}



- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	// catch rising edge of trigger
	if( [inputTrig wasUpdated] && [inputTrig booleanValue] )
		[self _startPlaying];


	if( [inputLoop wasUpdated] && ![inputLoop booleanValue] )
		for(NSSound *s in _allocatedSounds)
			[s setLoops:NO];

	if( [inputCurrentVolume wasUpdated] )
		for(NSSound *s in _allocatedSounds)
			[s setVolume:[inputCurrentVolume doubleValue]];

	if( [inputCurrentPosition wasUpdated] )
		for(NSSound *s in _allocatedSounds)
			[s setCurrentTime:[inputCurrentPosition doubleValue]];

	_executedSinceSetup=YES;

	return YES;
}






- (void)_startPlayingThread:(NSSound*)s
{
	[s play];
	[s release];
}

- (void)_startPlaying
{
	NSString *path = [inputPath stringValue];
	if([path length] > 0)
	{
		
		if([path characterAtIndex:0] != '/')
		{
			if([path characterAtIndex:0] != '~')
			{
				// not absolute and not ~-ified, must be composition relative
				NSDictionary *options = [self _getUserInfo:@".options"];
				NSURL *url = [options objectForKey:@"compositionURL"];
				path = [[[url path] stringByDeletingLastPathComponent] stringByAppendingPathComponent:path];
			}
			else    // it is a ~ directory, expand it
				path = [path stringByExpandingTildeInPath];
		}
	}


	NSSound *s=[[NSSound alloc] initWithContentsOfFile:path byReference:YES];
	if( !s )
	{
		[self logMessage:
			[NSString stringWithFormat:@"Failed to load audio file '%@'.",path]
		];
		return;
	}

//	NSLog(@"AudioFilePlayerPatch: %08x: alloc",s);

	[s setLoops:[inputLoop booleanValue]];
	[s setVolume:[inputCurrentVolume doubleValue]];
	[s setCurrentTime:[inputCurrentPosition doubleValue]];

	// pick up the sound:didFinishPlaying message so we can dealloc the NSSound
	[s setDelegate:self];

	// if a playback device isn't specified, use the default (by not forcing the playback device)
	if( [[inputDeviceUID stringValue] length] )
	{
		@try
		{
			[s setPlaybackDeviceIdentifier:[inputDeviceUID stringValue]];
		}
		@catch(...)
		{
			[self logMessage:
				[NSString stringWithFormat:@"Failed to set output device UID '%@'.",[inputDeviceUID stringValue]]
			];
			[s release];
			return;
		}
	}

	// if a channel mapping isn't specified, use the default (by not forcing the channel mapping)
	if( [[inputChannelMapping stringValue] length] )
	{
		NSMutableArray *channelMapping = [[NSMutableArray alloc] initWithCapacity:16];
		NSArray *channelMappingStrings = [[inputChannelMapping stringValue] componentsSeparatedByString:@","];
		for(NSString *s in channelMappingStrings)
			[channelMapping addObject:[NSNumber numberWithInteger:[s integerValue]]];
		[s setChannelMapping:channelMapping];
	}

	[NSThread detachNewThreadSelector:@selector(_startPlayingThread:) toTarget:self withObject:s];


	[_allocatedSounds addObject:s];
}


- (void)sound:(NSSound *)sound didFinishPlaying:(BOOL)finishedPlaying
{
//	NSLog(@"AudioFilePlayerPatch: %08x: release  (didFinishPlaying:%d)",sound,finishedPlaying);

	// this automatically releases the NSSound object
	[_allocatedSounds removeObject:sound];
}

@end
